from .clustering import *
from .conversion import GraphFormatConvertor
from .utils import add_labels_to_graph

try:
    from .datasets import (
        InMemoryProteinGraphDataset,
        ProteinGraphDataset,
        ProteinGraphListDataset,
    )
except (ImportError, NameError):
    pass
try:
    from .visualisation import *
except (ImportError, NameError):
    pass



"""Sequence featurisation functions wrapping ProPy."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Eric Ma
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

from functools import partial
from typing import Any, Callable, Dict, List, Optional

import networkx as nx
from loguru import logger as log

from graphein.protein.features.sequence.utils import (
    aggregate_feature_over_chains,
    compute_feature_over_chains,
)
from graphein.protein.features.utils import (
    aggregate_graph_feature_over_chains,
    convert_graph_dict_feat_to_series,
)


def compute_propy_feature(
    G: nx.Graph,
    func: Callable,
    feature_name: str,
    aggregation_type: Optional[List[str]] = None,
) -> nx.Graph:
    """
    Computes Propy Descriptors over chains in a Protein Graph

    :param G: Protein Graph
    :type G: nx.Graph
    :param func: ProPy wrapper function to compute
    :type func: Callable
    :param feature_name: Name of feature to index it in the nx.Graph object
    :type feature_name: str
    :param aggregation_type: Type of aggregation to use when aggregating a
        feature over multiple chains. One of: ``["mean", "max", "sum"]``.
        Defaults to ``None``.
    :type aggregation_type: List[str], optional
    :return G: Returns protein Graph with features added. Features are
        accessible with ``G.graph[{feature_name}_{chain | aggegation_type}]``
    :rtype: nx.Graph
    """
    G = compute_feature_over_chains(G, func, feature_name=feature_name)

    # Convert to Series
    G = convert_graph_dict_feat_to_series(G, feature_name=feature_name)

    # Aggregate features
    if len(G.graph["chain_ids"]) > 1:
        if aggregation_type:
            G = aggregate_graph_feature_over_chains(
                G, feature_name=feature_name, aggregation_type=aggregation_type
            )
    else:
        log.debug("Aggregation not carried out on single-chain graph")
    return G


def amino_acid_composition(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate the composition of Amino acids for a given protein sequence.

    :param G: Protein Graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use.
    :type aggregation_type: Optional[List[str]]
    :return: Protein Graph with ``amino_acid_composition`` feature added.
        ``G.graph["amino_acid_composition_{chain | aggregation_type}"]``
    :rtype: nx.Graph
    """
    from propy.AAComposition import CalculateAAComposition

    func = CalculateAAComposition
    feature_name = "amino_acid_composition"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def dipeptide_composition(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate the composition of dipeptide for a given protein sequence.
    Contains composition of 400 dipeptides.

    :param G: Protein Graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``dipeptide_composition`` feature added.
        ``G.graph["dipeptide_composition_{chain | aggregation_type}"]``
    :rtype: nx.Graph
    """
    from propy.AAComposition import CalculateDipeptideComposition

    func = CalculateDipeptideComposition
    feature_name = "dipeptide_composition"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def aa_dipeptide_composition(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate the composition of AADs, dipeptide and 3-mers for a given protein
    sequence. Contains all composition values of AADs, dipeptide and 3-mers
    (8420).

    :param G: Protein Graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``aa_dipeptide_composition`` feature added.
        ``G.graph["aa_dipeptide_composition_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.AAComposition import CalculateAADipeptideComposition

    func = CalculateAADipeptideComposition
    feature_name = "aa_dipeptide_composition"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def aa_spectrum(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate the spectrum descriptors of 3-mers for a given protein. Contains
    the composition values of 8000 3-mers

    :param G: Protein Graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with aa_spectrum feature added.
        ``G.graph["aa_spectrum_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.AAComposition import GetSpectrumDict

    func = GetSpectrumDict
    feature_name = "aa_spectrum"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


######## CTD Funcs #######


def all_composition_descriptors(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate all composition descriptors based on seven different properties of
    AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``composition_descriptors`` feature added.
        ``G.graph["composition_descriptors_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateC

    func = CalculateC
    feature_name = "composition_descriptors"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def all_ctd_descriptors(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate all CTD descriptors based seven different properties of AADs.

    :param G: Protein Graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``ctd_descriptors`` feature added.
        ``G.graph["ctd_descriptors_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateCTD

    func = CalculateCTD
    feature_name = "ctd_descriptors"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def composition_descriptor(
    G: nx.Graph,
    AAProperty: Dict[Any, Any],
    AAPName: str,
    aggregation_type: Optional[List[str]] = None,
) -> nx.Graph:
    """
    Compute composition descriptors.

    :param G: Protein Graph to featurise.
    :type G: nx.Graph
    :param AAProperty: contains classification of amino acids such as
        ``_Polarizability.``
    :type AAProperty: Dict[Any, Any]
    :param AAPName: used for indicating a AAP name.
    :type AAPName: str
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``composition_{AAPName}`` feature added.
        ``G.graph["composition_{AAPName}_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateComposition

    func = partial(
        CalculateComposition, AAProperty=AAProperty, AAPName=AAPName
    )
    feature_name = f"composition_{AAPName}"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def composition_charge(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate composition descriptors based on Charge of AADs.

    :param G: Protein Graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``composition_charge`` feature added.
        ``G.graph["composition_charge_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateCompositionCharge

    func = CalculateCompositionCharge
    feature_name = "composition_charge"

    G = compute_feature_over_chains(G, func, feature_name="composition_charge")

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def composition_hydrophobicity(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate composition descriptors based on Hydrophobicity of AADs.

    :param G: Protein Graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``composition_hydrophobicity`` feature added.
        ``G.graph["composition_hydrophobicity_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateCompositionHydrophobicity

    func = CalculateCompositionHydrophobicity
    feature_name = "composition_hydrophobicity"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def composition_normalized_vdwv(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate composition descriptors based on NormalizedVDWV of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``composition_normalized_vdwv`` feature added.
        ``G.graph["composition_normalized_vdwv_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateCompositionNormalizedVDWV

    func = CalculateCompositionNormalizedVDWV
    feature_name = "composition_normalised_vdwv"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def composition_polarity(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate composition descriptors based on Polarity of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``composition_polarity`` feature added.
        ``G.graph["composition_polarity_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateCompositionPolarity

    func = CalculateCompositionPolarity
    feature_name = "composition_polarity"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def composition_polarizability(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate composition descriptors based on Polarizability of AADs.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``composition_polarizability`` feature added.
        ``G.graph["composition_polarizability_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateCompositionPolarizability

    func = CalculateCompositionPolarizability
    feature_name = "composition_polarizability"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def composition_secondary_str(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate composition descriptors based on SecondaryStr of AADs.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``composition_secondary_str`` feature added.
        ``G.graph["composition_secondary_str_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateCompositionSecondaryStr

    func = CalculateCompositionSecondaryStr
    feature_name = "composition_secondary_str"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def composition_solvent_accessibility(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate composition descriptors based on SolventAccessibility of AADs.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with composition_solvent_accessibility feature added.
        G.graph["composition_solvent_accessibility_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateCompositionSolventAccessibility

    func = CalculateCompositionSolventAccessibility
    feature_name = "composition_solvent_accessibility"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def all_distribution_descriptors(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate all distribution descriptors based on seven different properties
    of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``distribution_descriptors`` feature added.
        ``G.graph["distribution_descriptors_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateC

    func = CalculateC
    feature_name = "distribution_descriptors"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def distribution_descriptor(
    G: nx.Graph,
    AAProperty: Dict[Any, Any],
    AAPName: str,
    aggregation_type: Optional[List[str]] = None,
) -> nx.Graph:
    """
    Compute distribution descriptors.

    :param G: Protein Graph to featurise.
    :type G: nx.Graph
    :param AAProperty: contains classifciation of amino acids such as _Polarizability.
    :type AAProperty: Dict[Any, Any]
    :param AAPName: used for indicating a AAP name.
    :type AAPName: str
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``distribution_{AAPName}`` feature added.
        ``G.graph["distribution_{AAPName}_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateDistribution

    func = partial(
        CalculateDistribution, AAProperty=AAProperty, AAPName=AAPName
    )
    feature_name = f"distribution_{AAPName}"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def distribution_charge(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate distribution descriptors based on Charge of AADs.

    :param G: Protein Graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``distribution_charge`` feature added.
        ``G.graph["distribution_charge_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateDistributionCharge

    func = CalculateDistributionCharge
    feature_name = "distribution_charge"

    G = compute_feature_over_chains(
        G, func, feature_name="distribution_charge"
    )

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def distribution_hydrophobicity(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate distribution descriptors based on Hydrophobicity of AADs.

    :param G: Protein Graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``distribution_hydrophobicity`` feature added.
        ``G.graph["distribution_hydrophobicity_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateDistributionHydrophobicity

    func = CalculateDistributionHydrophobicity
    feature_name = "distribution_hydrophobicity"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def distribution_normalized_vdwv(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate distribution descriptors based on NormalizedVDWV of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``distribution_normalized_vdwv`` feature added.
        ``G.graph["distribution_normalized_vdwv_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateDistributionNormalizedVDWV

    func = CalculateDistributionNormalizedVDWV
    feature_name = "distribution_normalised_vdwv"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def distribution_polarity(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate distribution descriptors based on Polarity of AADs.

    :param G: Protein Graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``distribution_polarity`` feature added.
        ``G.graph["distribution_polarity_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateDistributionPolarity

    func = CalculateDistributionPolarity
    feature_name = "distribution_polarity"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def distribution_polarizability(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate distribution descriptors based on Polarizability of AADs.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``distribution_polarizability`` feature added.
        ``G.graph["distribution_polarizability_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateDistributionPolarizability

    func = CalculateDistributionPolarizability
    feature_name = "distribution_polarizability"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def distribution_secondary_str(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate distribution descriptors based on SecondaryStr of AADs.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``distribution_secondary_str`` feature added.
        ``G.graph["distribution_secondary_str_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateDistributionSecondaryStr

    func = CalculateDistributionSecondaryStr
    feature_name = "distribution_secondary_str"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def distribution_solvent_accessibility(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate distribution descriptors based on SolventAccessibility of AADs.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``distribution_solvent_accessibility`` feature
        added.
        ``G.graph["distribution_solvent_accessibility_{chain | aggregation_type}"]``
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateDistributionSolventAccessibility

    func = CalculateDistributionSolventAccessibility
    feature_name = "distribution_solvent_accessibility"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def all_transition_descriptors(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate all transition descriptors based on seven different properties of
    AADs.

    :param G: Protein Graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``transition_descriptors`` feature added.
        ``G.graph["transition_descriptors_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateC

    func = CalculateC
    feature_name = "transition_descriptors"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def transition_descriptor(
    G: nx.Graph,
    AAProperty: Dict[Any, Any],
    AAPName: str,
    aggregation_type: Optional[List[str]] = None,
) -> nx.Graph:
    """
    Compute transition descriptors.

    :param G: Protein Graph to featurise.
    :type G: nx.Graph
    :param AAProperty: contains classifciation of amino acids such as _Polarizability.
    :type AAProperty: Dict[Any, Any]
    :param AAPName: used for indicating a AAP name.
    :type AAPName: str
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``transition_{AAPName}`` feature added.
        ``G.graph["transition_{AAPName}_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateTransition

    func = partial(CalculateTransition, AAProperty=AAProperty, AAPName=AAPName)
    feature_name = f"transition_{AAPName}"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def transition_charge(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate transition descriptors based on Charge of AADs.

    :param G: Protein Graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``transition_charge`` feature added.
        ``G.graph["transition_charge_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateTransitionCharge

    func = CalculateTransitionCharge
    feature_name = "transition_charge"

    G = compute_feature_over_chains(G, func, feature_name="transition_charge")

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def transition_hydrophobicity(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate transition descriptors based on Hydrophobicity of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``transition_hydrophobicity`` feature added.
        ``G.graph["transition_hydrophobicity_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateTransitionHydrophobicity

    func = CalculateTransitionHydrophobicity
    feature_name = "transition_hydrophobicity"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def transition_normalized_vdwv(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate transition descriptors based on NormalizedVDWV of AADs.

    :param G: Protein Graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``transition_normalized_vdwv`` feature added.
        ``G.graph["transition_normalized_vdwv_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateTransitionNormalizedVDWV

    func = CalculateTransitionNormalizedVDWV
    feature_name = "transition_normalised_vdwv"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def transition_polarity(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate transition descriptors based on Polarity of AADs.

    :param G: Protein Graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``transition_polarity`` feature added.
        ``G.graph["transition_polarity_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateTransitionPolarity

    func = CalculateTransitionPolarity
    feature_name = "transition_polarity"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def transition_polarizability(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate transition descriptors based on Polarizability of AADs.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``transition_polarizability`` feature added.
        ``G.graph["transition_polarizability_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateTransitionPolarizability

    func = CalculateTransitionPolarizability
    feature_name = "transition_polarizability"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def transition_secondary_str(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate transition descriptors based on SecondaryStr of AADs.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``transition_secondary_str`` feature added.
        ``G.graph["transition_secondary_str_{chain | aggregation_type}"]``
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateTransitionSecondaryStr

    func = CalculateTransitionSecondaryStr
    feature_name = "transition_secondary_str"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def transition_solvent_accessibility(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate transition descriptors based on SolventAccessibility of AADs.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``transition_solvent_accessibility`` feature
        added.
        ``G.graph["transition_solvent_accessibility_{chain | aggregation_type}"]``
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateTransitionSolventAccessibility

    func = CalculateTransitionSolventAccessibility
    feature_name = "transition_solvent_accessibility"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


####### Autocorrelation descriptors ########


def autocorrelation_total(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Compute all autocorrelation descriptors based on 8 properties of AADs.
    result contains 30*8*3=720 normalized Moreau Broto, Moran, and Geary
    autocorrelation descriptors.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_total`` feature added.
        ``G.graph["autocorrelation_total_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateAutoTotal

    func = CalculateAutoTotal
    feature_name = "autocorrelation_total"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_geary_all(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Compute Geary autocorrelation descriptors based on 8 properties of AADs.
    Result contains 30*8=240 Geary autocorrelation descriptors based on the
    given properties(i.e., _AAPropert).

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_geary_all`` feature added.
        ``G.graph["autocorrelation_geary_all_{chain | aggregation_type}"]``
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateGearyAutoTotal

    func = CalculateGearyAutoTotal
    feature_name = "autocorrelation_geary_all"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_geary_av_flexibility(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Geary Autocorrelation descriptors based on AvFlexibility.
    contains 30 Geary Autocorrelation descriptors based on AvFlexibility.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_geary_av_flexibility``
        feature added.
        ``G.graph["autocorrelation_geary_av_flexibility_{chain | aggregation_type}"]``
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateGearyAutoAvFlexibility

    func = CalculateGearyAutoAvFlexibility
    feature_name = "autocorrelation_geary_av_flexibility"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_geary_free_energy(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Geary Autocorrelation descriptors based on FreeEnergy.
    Result contains 30 Geary Autocorrelation descriptors based on FreeEnergy.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_geary_free_energy`` feature
        added.
        ``G.graph["autocorrelation_geary_av_free_energy_{chain | aggregation_type}"]``
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateGearyAutoFreeEnergy

    func = CalculateGearyAutoFreeEnergy
    feature_name = "autocorrelation_geary_free_energy"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_geary_hydrophobicity(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Geary Autocorrelation descriptors based on hydrophobicity.
    result contains 30 Geary Autocorrelation descriptors based on hydrophobicity.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_geary_hydrophobicity`` feature
        added.
        ``G.graph["autocorrelation_geary_hydrophobicity_{chain | aggregation_type}"]``
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateGearyAutoHydrophobicity

    func = CalculateGearyAutoHydrophobicity
    feature_name = "autocorrelation_geary_hydrophobicity"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_geary_mutability(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Geary Autocorrelation descriptors based on Mutability.
    Result contains 30 Geary Autocorrelation descriptors based on mutability.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_geary_mutability`` feature
        added.
        ``G.graph["autocorrelation_geary_mutability_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateGearyAutoMutability

    func = CalculateGearyAutoMutability
    feature_name = "autocorrelation_geary_mutability"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_geary_polarizability(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Geary Autocorrelation descriptors based on polarizability.
    Result contains 30 Geary Autocorrelation descriptors based on
    polarizability.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_geary_polarizability``
        feature added.
        ``G.graph["autocorrelation_geary_polarizability_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateGearyAutoPolarizability

    func = CalculateGearyAutoPolarizability
    feature_name = "autocorrelation_geary_polarizability"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_geary_residue_asa(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Geary Autocorrelation descriptors based on ResidueASA.
    Result contains 30 Geary Autocorrelation descriptors based on ResidueASA.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_geary_residue_asa`` feature added.
        ``G.graph["autocorrelation_geary_residue_asa_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateGearyAutoResidueASA

    func = CalculateGearyAutoResidueASA
    feature_name = "autocorrelation_geary_residue_asa"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_geary_residue_vol(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Geary Autocorrelation descriptors based on ResidueVol.
    Result contains 30 Geary Autocorrelation descriptors based on ResidueVol.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_geary_residue_vol`` feature
        added.
        ``G.graph["autocorrelation_geary_residue_vol_{chain | aggregation_type}"]``
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateGearyAutoResidueVol

    func = CalculateGearyAutoResidueVol
    feature_name = "autocorrelation_geary_residue_vol"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_geary_steric(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Geary Autocorrelation descriptors based on Steric. Result
    contains 30 Geary Autocorrelation descriptors based on Steric

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_geary_steric`` feature added.
        ``G.graph["autocorrelation_geary_steric_{chain | aggregation_type}"]``
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateGearyAutoSteric

    func = CalculateGearyAutoSteric()
    feature_name = "autocorrelation_geary_steric"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_moran_all(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Compute Moran autocorrelation descriptors based on 8 properties of AADs.
    Result contains 30*8=240 Moran autocorrelation descriptors based on the
    given properties(i.e., _AAPropert).

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_moran_all`` feature added.
        ``G.graph["autocorrelation_moran_all_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateMoranAutoTotal

    func = CalculateMoranAutoTotal
    feature_name = "autocorrelation_moran_all"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_moran_av_flexibility(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Moran Autocorrelation descriptors based on AvFlexibility.
    Contains 30 Moran Autocorrelation descriptors based on AvFlexibility.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_moran_av_flexibility``
        feature added.
        ``G.graph["autocorrelation_moran_av_flexibility_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateMoranAutoAvFlexibility

    func = CalculateMoranAutoAvFlexibility
    feature_name = "autocorrelation_moran_av_flexibility"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_moran_free_energy(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Moran Autocorrelation descriptors based on FreeEnergy.
    Result contains 30 Moran Autocorrelation descriptors based on FreeEnergy.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_moran_free_energy`` feature
        added.
        ``G.graph["autocorrelation_moran_av_free_energy_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateMoranAutoFreeEnergy

    func = CalculateMoranAutoFreeEnergy
    feature_name = "autocorrelation_moran_free_energy"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_moran_hydrophobicity(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Moran Autocorrelation descriptors based on hydrophobicity.
    Result contains 30 Moran Autocorrelation descriptors based on hydrophobicity.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_moran_hydrophobicity`` feature
        added.
        ``G.graph["autocorrelation_moran_hydrophobicity_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateMoranAutoHydrophobicity

    func = CalculateMoranAutoHydrophobicity
    feature_name = "autocorrelation_moran_hydrophobicity"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_moran_mutability(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Moran Autocorrelation descriptors based on Mutability.
    Result contains 30 Moran Autocorrelation descriptors based on mutability.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_moran_mutability`` feature
        added.
        ``G.graph["autocorrelation_moran_mutability_{chain | aggregation_type}"]``
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateMoranAutoMutability

    func = CalculateMoranAutoMutability
    feature_name = "autocorrelation_moran_mutability"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_moran_polarizability(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Moran Autocorrelation descriptors based on polarizability.
    Result contains 30 Moran Autocorrelation descriptors based on polarizability.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_moran_polarizability`` feature
        added.
        ``G.graph["autocorrelation_moran_polarizability_{chain | aggregation_type}"]``
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateMoranAutoPolarizability

    func = CalculateMoranAutoPolarizability
    feature_name = "autocorrelation_moran_polarizability"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_moran_residue_asa(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Moran Autocorrelation descriptors based on ResidueASA.
    Result contains 30 Moran Autocorrelation descriptors based on ResidueASA.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_moran_residue_asa`` feature added.
        ``G.graph["autocorrelation_moran_residue_asa_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateMoranAutoResidueASA

    func = CalculateMoranAutoResidueASA
    feature_name = "autocorrelation_moran_residue_asa"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_moran_residue_vol(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Moran Autocorrelation descriptors based on ResidueVol.
    Result contains 30 Moran Autocorrelation descriptors based on ResidueVol.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_moran_residue_vol`` feature
        added.
        ``G.graph["autocorrelation_moran_residue_vol_{chain | aggregation_type}"]``
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateMoranAutoResidueVol

    func = CalculateMoranAutoResidueVol
    feature_name = "autocorrelation_moran_residue_vol"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_moran_steric(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Moran Autocorrelation descriptors based on Steric. Result
    contains 30 Moran Autocorrelation descriptors based on Steric

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_moran_steric`` feature added.
        ``G.graph["autocorrelation_moran_steric_{chain | aggregation_type}"]``
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateMoranAutoSteric

    func = CalculateMoranAutoSteric()
    feature_name = "autocorrelation_moran_steric"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_normalized_moreau_broto_all(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Compute NormalizedMoreauBroto autocorrelation descriptors based on 8
    properties of AADs. Result contains 30*8=240 NormalizedMoreauBroto
    autocorrelation descriptors based on the given properties(i.e., _AAPropert).

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ``autocorrelation_normalized_moreau_broto_all``
        feature added.
        ``G.graph["autocorrelation_normalized_moreau_broto_all_{chain | aggregation_type}"]``
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateNormalizedMoreauBrotoAutoTotal

    func = CalculateNormalizedMoreauBrotoAutoTotal
    feature_name = "autocorrelation_normalized_moreau_broto_all"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_normalized_moreau_broto_av_flexibility(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the NormalizedMoreauBroto Autocorrelation descriptors based on
    AvFlexibility. contains 30 NormalizedMoreauBroto Autocorrelation descriptors
    based on AvFlexibility.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with
        ``autocorrelation_normalized_moreau_broto_av_flexibility`` feature
        added.
        ``G.graph["autocorrelation_normalized_moreau_broto_av_flexibility_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import (
        CalculateNormalizedMoreauBrotoAutoAvFlexibility,
    )

    func = CalculateNormalizedMoreauBrotoAutoAvFlexibility
    feature_name = "autocorrelation_normalized_moreau_broto_av_flexibility"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_normalized_moreau_broto_free_energy(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the NormalizedMoreauBroto Autocorrelation descriptors based on
    FreeEnergy. Result contains 30 NormalizedMoreauBroto Autocorrelation
    descriptors based on FreeEnergy.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with
        ``autocorrelation_normalized_moreau_broto_free_energy`` feature added.
        ``G.graph["autocorrelation_normalized_moreau_broto_av_free_energy_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import (
        CalculateNormalizedMoreauBrotoAutoFreeEnergy,
    )

    func = CalculateNormalizedMoreauBrotoAutoFreeEnergy
    feature_name = "autocorrelation_normalized_moreau_broto_free_energy"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_normalized_moreau_broto_hydrophobicity(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the NormalizedMoreauBroto Autocorrelation descriptors based on
    hydrophobicity. Result contains 30 NormalizedMoreauBroto Autocorrelation
    descriptors based on hydrophobicity.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with
        ``autocorrelation_normalized_moreau_broto_hydrophobicity`` feature
        added.
        ``G.graph["autocorrelation_normalized_moreau_broto_hydrophobicity_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import (
        CalculateNormalizedMoreauBrotoAutoHydrophobicity,
    )

    func = CalculateNormalizedMoreauBrotoAutoHydrophobicity
    feature_name = "autocorrelation_normalized_moreau_broto_hydrophobicity"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_normalized_moreau_broto_mutability(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the NormalizedMoreauBroto Autocorrelation descriptors based on
    Mutability. Result contains 30 NormalizedMoreauBroto Autocorrelation
    descriptors based on mutability.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with
        ``autocorrelation_normalized_moreau_broto_mutability`` feature added.
        ``G.graph["autocorrelation_normalized_moreau_broto_mutability_{chain | aggregation_type}"]``
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import (
        CalculateNormalizedMoreauBrotoAutoMutability,
    )

    func = CalculateNormalizedMoreauBrotoAutoMutability
    feature_name = "autocorrelation_normalized_moreau_broto_mutability"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_normalized_moreau_broto_polarizability(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the NormalizedMoreauBroto Autocorrelation descriptors based on
    polarizability. Result contains 30 NormalizedMoreauBroto Autocorrelation
    descriptors based on polarizability.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with
        ``autocorrelation_normalized_moreau_broto_polarizability`` feature
        added.
        ``G.graph["autocorrelation_normalized_moreau_broto_polarizability_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import (
        CalculateNormalizedMoreauBrotoAutoPolarizability,
    )

    func = CalculateNormalizedMoreauBrotoAutoPolarizability
    feature_name = "autocorrelation_normalized_moreau_broto_polarizability"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_normalized_moreau_broto_residue_asa(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the NormalizedMoreauBroto Autocorrelation descriptors based on
    ResidueASA. Result contains 30 NormalizedMoreauBroto Autocorrelation
    descriptors based on ResidueASA.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with
        ``autocorrelation_normalized_moreau_broto_residue_asa`` feature added.
        ``G.graph["autocorrelation_normalized_moreau_broto_residue_asa_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import (
        CalculateNormalizedMoreauBrotoAutoResidueASA,
    )

    func = CalculateNormalizedMoreauBrotoAutoResidueASA
    feature_name = "autocorrelation_normalized_moreau_broto_residue_asa"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_normalized_moreau_broto_residue_vol(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the NormalizedMoreauBroto Autocorrelation descriptors based on
    ResidueVol. Result contains 30 NormalizedMoreauBroto Autocorrelation
    descriptors based on ResidueVol.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with
        ``autocorrelation_normalized_moreau_broto_residue_vol`` feature added.
        ``G.graph["autocorrelation_normalized_moreau_broto_residue_vol_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import (
        CalculateNormalizedMoreauBrotoAutoResidueVol,
    )

    func = CalculateNormalizedMoreauBrotoAutoResidueVol
    feature_name = "autocorrelation_normalized_moreau_broto_residue_vol"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_normalized_moreau_broto_steric(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the NormalizedMoreauBroto Autocorrelation descriptors based on
    Steric. Result contains 30 NormalizedMoreauBroto Autocorrelation descriptors
    based on Steric.

    :param G: Protein graph to featurise.
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains.
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with
        ``autocorrelation_normalized_moreau_broto_steric`` feature added.
        ``G.graph["autocorrelation_normalized_moreau_broto_steric_{chain | aggregation_type}"]``.
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateNormalizedMoreauBrotoAutoSteric

    func = CalculateNormalizedMoreauBrotoAutoSteric()
    feature_name = "autocorrelation_normalized_moreau_broto_steric"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


####### Quasi Sequence Order Descriptors ######


def quasi_sequence_order_aa_composition(G: nx.Graph) -> nx.Graph:
    from propy.QuasiSequenceOrder import GetAAComposition

    func = GetAAComposition
    G = compute_feature_over_chains(
        G, func, feature_name="quasi_seq_order_aa_composition"
    )

    return G


def quasi_sequence_order(
    G: nx.Graph, maxlag: int = 30, weight: float = 0.1
) -> nx.Graph:
    """
    Compute quasi-sequence-order descriptors for a given protein.

    Kuo-Chen Chou. Prediction of Protein Subcellar Locations by Incorporating
    Quasi-Sequence-Order Effect. Biochemical and Biophysical Research
    Communications 2000, 278, 477-483.

    :param maxlag: The maximum lag and the
        length of the protein should be larger than maxlag.
    :type maxlag: int, defaults to ``30``.
    :param weight: Weight factor. Please see reference 1 for its choice.
    :type weight: float, defaults to ``0.1``.
    :returns: Protein Graph with ``quasi_sequence_order`` feature added.
    :rtype: nx.Graph
    """
    from propy.QuasiSequenceOrder import GetQuasiSequenceOrder

    func = partial(GetQuasiSequenceOrder, weight=weight, maxlag=maxlag)
    G = compute_feature_over_chains(
        G, func, feature_name="quasi_sequence_order"
    )
    return G


def sequence_order_coupling_number_total(
    G: nx.Graph, maxlag: int = 30
) -> nx.Graph:
    """
    Compute the sequence order coupling numbers from 1 to maxlag for a given
    protein sequence.

    :param G: Protein Graph.
    :type G: nx.Graph
    :param maxlag: The maximum lag and the length of the protein should be
        larger.
    :type maxlag: int, defaults to ``30``.
    :returns: Protein Graph with ``sequence_order_coupling_number_total``
        feature.
    :rtype: nx.Graph
    """
    from propy.QuasiSequenceOrder import GetSequenceOrderCouplingNumberTotal

    func = partial(GetSequenceOrderCouplingNumberTotal, maxlag=maxlag)
    G = compute_feature_over_chains(
        G, func, feature_name="sequence_order_coupling_number_total"
    )

    return G


# TODO feature aggregation



"""Featurization functions for amino acids."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Eric Ma
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd
from loguru import logger as log

from graphein.protein.resi_atoms import (
    BASE_AMINO_ACIDS,
    HYDROGEN_BOND_ACCEPTORS,
    HYDROGEN_BOND_DONORS,
    RESI_THREE_TO_1,
)
from graphein.utils.utils import onek_encoding_unk


@lru_cache()
def load_expasy_scales() -> pd.DataFrame:
    """
    Load pre-downloaded EXPASY scales.

    This helps with node featuarization.

    The function is LRU-cached in memory for fast access
    on each function call.

    :returns: pd.DataFrame containing expasy scales
    :rtype: pd.DataFrame
    """
    fpath = Path(__file__).parent / "amino_acid_properties.csv"
    log.debug(f"Reading Expasy protein scales from: {fpath}")
    return pd.read_csv(fpath, index_col=0)


@lru_cache()
def load_meiler_embeddings() -> pd.DataFrame:
    """
    Load pre-downloaded Meiler embeddings.

    This helps with node featurization.

    The function is LRU-cached in memory for fast access
    on each function call.

    :returns: pd.DataFrame containing Meiler Embeddings from Meiler et al. 2001
    :rtype: pd.DataFrame
    """
    fpath = Path(__file__).parent / "meiler_embeddings.csv"
    log.debug(f"Reading meiler embeddings from: {fpath}")
    return pd.read_csv(fpath, index_col=0)


def expasy_protein_scale(
    n: str,
    d: Dict[str, Any],
    selection: Optional[List[str]] = None,
    add_separate: bool = False,
    return_array: bool = False,
) -> Union[pd.Series, np.ndarray]:
    """
    Return amino acid features that come from the EXPASY protein scale.

    Source: https://web.expasy.org/protscale/

    :param n: Node in a NetworkX graph
    :type n: str
    :param d: NetworkX node attributes.
    :type d: Dict[str, Any]
    :param selection: List of columns to select. Viewable in
        :ref:`~graphein.protein.features.nodes.meiler_embeddings`.
    :type selection: List[str], optional
    :param add_separate: Whether or not to add the expasy features as individual
        entries or as a series.
    :param return_array: Bool indicating whether or not to return a
        ``np.ndarray`` of the features. Default is ``pd.Series``.
    :type return_array: bool
    :returns: pd.Series of amino acid features
    :rtype: pd.Series
    """
    df = load_expasy_scales()
    amino_acid = d["residue_name"]
    try:
        features = df[amino_acid]
        if selection is not None:
            features = features.filter(selection)
    except:
        features = pd.Series(np.zeros(len(df)))

    if return_array:
        features = np.array(features)

    if add_separate:
        for k, v in features.to_dict().items():
            d[k] = v
    else:
        d["expasy"] = features

    return features


def meiler_embedding(
    n: str, d: Dict[str, Any], return_array: bool = False
) -> Union[pd.Series, np.ndarray]:
    """
    Return amino acid features from reduced dimensional embeddings of amino
    acid physicochemical properties.

    Source: https://link.springer.com/article/10.1007/s008940100038
    doi: https://doi.org/10.1007/s008940100038

    :param n: Node in a NetworkX graph
    :type n: str
    :param d: NetworkX node attributes.
    :type d: Dict[str, Any]
    :param return_array: Bool indicating whether or not to return a
        ``np.ndarray`` of the features. Default is ``pd.Series``.
    :returns: pd.Series of amino acid features
    :rtype: pd.Series
    """
    df = load_meiler_embeddings()
    amino_acid = d["residue_name"]
    try:
        features = df[amino_acid]
    except KeyError:
        features = pd.Series(np.zeros(len(df)))

    if return_array:
        features = np.array(features)

    d["meiler"] = features

    return features


def amino_acid_one_hot(
    n,
    d: Dict[str, Any],
    return_array: bool = True,
    allowable_set: Optional[List[str]] = None,
) -> Union[pd.Series, np.ndarray]:
    """Adds a one-hot encoding of amino acid types as a node attribute.

    :param n: node name, this is unused and only included for compatibility
        with the other functions
    :type n: str
    :param d: Node data.
    :type d: Dict[str, Any]
    :param return_array: If True, returns a numpy array of one-hot encoding,
        otherwise returns a pd.Series. Default is True.
    :type return_array: bool
    :param allowable_set: Specifies vocabulary of amino acids. Default is
        ``None`` (which uses
        :const:`~graphein.protein.resi_atoms.STANDARD_AMINO_ACIDS`).
    :return: One-hot encoding of amino acid types.
    :rtype: Union[pd.Series, np.ndarray]
    """

    if allowable_set is None:
        allowable_set = BASE_AMINO_ACIDS

    features = onek_encoding_unk(
        RESI_THREE_TO_1[d["residue_name"]], allowable_set
    )

    if return_array:
        features = np.array(features).astype(int)
    else:
        features = pd.Series(features).astype(int)
        features.index = allowable_set

    d["amino_acid_one_hot"] = features
    return features


def hydrogen_bond_donor(
    n: str,
    d: Dict[str, Any],
    sum_features: bool = True,
    return_array: bool = False,
):
    """Adds Hydrogen Bond Donor status to nodes as a feature.

    :param n: Node id
    :type n: str
    :param d: Dict of node attributes
    :type d: Dict[str, Any]
    :param sum_features: If ``True``, the feature is the number of hydrogen
        bond donors per node. If ``False``, the feature is a boolean indicating
        whether or not the node has a hydrogen bond donor. Default is ``True``.
    :type sum_features: bool
    :param return_array: If ``True``, returns a ``np.ndarray``, otherwise
        returns a ``pd.Series``. Default is ``True``.
    :type return_array: bool
    """
    res = d["residue_name"]

    # Hack to determine graph type
    # If last ID component is atom type, assume graph is atomic
    if n.split(":")[-1] == d["atom_type"]:
        granularity = "atom"
    else:
        granularity = "residue"

    if granularity == "atom":  # Atomic graph
        atom = d["atom_type"]
        try:
            features = HYDROGEN_BOND_DONORS[res][atom]
        except KeyError:
            features = 0
    else:  # Residue graph
        if res not in HYDROGEN_BOND_DONORS.keys():
            features = 0
        else:
            features = sum(HYDROGEN_BOND_DONORS[res].values())

    if return_array:
        features = np.array(features).astype(int)
    else:
        features = pd.Series(features).astype(int)
    if not sum_features:
        features = np.array(features > 0).astype(int)

    d["hbond_donors"] = features


def hydrogen_bond_acceptor(
    n, d, sum_features: bool = True, return_array: bool = False
) -> pd.Series:
    """Adds Hydrogen Bond Acceptor status to nodes as a feature.

    :param n: Node id
    :type n: str
    :param d: Dict of node attributes.
    :type d: Dict[str, Any]
    :param sum_features: If ``True``, the feature is the number of hydrogen
        bond acceptors per node. If ``False``, the feature is a boolean
        indicating whether or not the node has a hydrogen bond acceptor.
        Default is ``True``.
    :type sum_features: bool
    :param return_array: If ``True``, returns a ``np.ndarray``, otherwise
        returns a ``pd.Series``. Default is ``True``.
    :type return_array: bool
    """
    res = d["residue_name"]
    # Hack to determine graph type
    # If last ID component is atom type, assume graph is atomic
    if n.split(":")[-1] == d["atom_type"]:
        granularity = "atom"
    else:
        granularity = "residue"

    if granularity == "atom":  # Atomic graph
        atom = d["atom_type"]
        try:
            features = HYDROGEN_BOND_ACCEPTORS[res][atom]
        except KeyError:
            features = 0
    else:  # Residue graph
        if res not in HYDROGEN_BOND_ACCEPTORS.keys():
            features = 0
        else:
            features = sum(HYDROGEN_BOND_ACCEPTORS[res].values())

    if return_array:
        features = np.array(features).astype(int)
    else:
        features = pd.Series(features).astype(int)
    if not sum_features:
        features = np.array(features > 0).astype(int)
    d["hbond_acceptors"] = features



from graphein.rna.nussinov import nussinov

TEST_SEQUENCE = "UUGGAGUACACAACCUGUACACUCUUUC"


def test_nussinov():
    # nussinov algo does not guarantee that the dot-bracket notation is correct.
    # There are several other ways of computing this.
    target_dot_bracket = nussinov(TEST_SEQUENCE)
    assert len(TEST_SEQUENCE) == len(target_dot_bracket)



"""Functions for computing atomic structure of proteins."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

from typing import Any, Dict

import networkx as nx
import numpy as np
from loguru import logger as log
from scipy.spatial import distance_matrix

import graphein.protein.edges.atomic as protein
from graphein.rna.features.atom import add_atomic_radii

# TODO dealing with metals
# TODO There are other check and balances that can be implemented from here:
# https://www.daylight.com/meetings/mug01/Sayle/m4xbondage.html


def add_atomic_edges(G: nx.Graph, tolerance: float = 0.56) -> nx.Graph:
    """
    Computes covalent edges based on atomic distances. Covalent radii are
    assigned to each atom based on its bond state. The distance matrix is then
    thresholded to entries less than this distance plus some tolerance to
    create an adjacency matrix. This adjacency matrix is then parsed into an
    edge list and covalent edges added.

    Bond states and covalent radii are retrieved from:

        Structures of the Molecular Components in DNA and RNA with Bond Lengths
        Interpreted as Sums of Atomic Covalent Radii
        *Raji Heyrovska*

    :param G: Atomic graph (nodes correspond to atoms) to populate with atomic
        bonds as edges
    :type G: nx.Graph
    :param tolerance: Tolerance for atomic distance. Default is ``0.56``
        Angstroms. Commonly used values are: ``0.4, 0.45, 0.56``
    :type tolerance: float
    :return: Atomic graph with edges between bonded atoms added
    :rtype: nx.Graph
    """
    dist_mat = distance_matrix(G.graph["coords"], G.graph["coords"])

    # Add atomic radii if not present
    for n, d in G.nodes(data=True):
        if "atomic_radius" not in d.keys():
            add_atomic_radii(n, d)

    radii = np.array(list(nx.get_node_attributes(G, "atomic_radius").values()))

    # Create a covalent 'distance' matrix by adding the radius arrays with
    # its transpose
    covalent_radius_distance_matrix = np.add(
        radii.reshape(-1, 1),
        radii.reshape(1, -1),
    )
    # Add the tolerance
    covalent_radius_distance_matrix = (
        covalent_radius_distance_matrix + tolerance
    )

    # Threshold Distance Matrix to entries where the euclidean distance is
    # less than the covalent radius plus tolerance and larger than 0.4
    dist_mat[dist_mat < 0.4] = np.nan
    dist_mat[dist_mat > covalent_radius_distance_matrix] = np.nan
    # Store atomic adjacency matrix in graph
    G.graph["atomic_adj_mat"] = np.nan_to_num(dist_mat)

    # Get node IDs from non NaN entries in the thresholded distance matrix
    # and add the edge to the graph
    inds = zip(*np.where(~np.isnan(dist_mat)))
    for i in inds:
        length = dist_mat[i[0]][i[1]]
        node_1 = G.graph["pdb_df"]["node_id"][i[0]]
        node_2 = G.graph["pdb_df"]["node_id"][i[1]]
        chain_1 = G.graph["pdb_df"]["chain_id"][i[0]]
        chain_2 = G.graph["pdb_df"]["chain_id"][i[1]]

        # Check nodes are in graph
        if not (G.has_node(node_1) and G.has_node(node_2)):
            continue

        # Check atoms are in the same chain
        if not (chain_1 and chain_2):
            continue

        if G.has_edge(node_1, node_2):
            G.edges[node_1, node_2]["kind"].add("covalent")
            G.edges[node_1, node_2]["bond_length"] = length
        else:
            G.add_edge(node_1, node_2, kind={"covalent"}, bond_length=length)

    # TODO checking degree against MAX_NEIGHBOURS

    return G


def add_ring_status(G: nx.Graph) -> nx.Graph:
    """
    Identifies rings in the atomic RNA graph. Assigns the edge attribute
    ``"RING"`` to edges in the ring. We do not distinguish between aromatic and
    non-aromatic rings. Functions by identifying all cycles in the graph.

    :param G: Atom-level RNA structure graph to add ring edge types to
    :type G: nx.Graph
    :return: Atom-level RNA structure graph with added ``"RING"`` edge attribute
    :rtype: nx.Graph
    """
    return protein.add_ring_status(G)


def add_bond_order(G: nx.Graph) -> nx.Graph:
    """
    Assign bond orders to the covalent bond edges between atoms on the basis of
    bond length. Values are taken from:

        Automatic Assignment of Chemical Connectivity to Organic Molecules in
        the Cambridge Structural Database.
        *Jon C. Baber and Edward E. Hodgkin*

    :param G: Atomic-level RNA graph with covalent edges.
    :type G: nx.Graph
    :return: Atomic-level RNA graph with covalent edges annotated with putative
        bond order.
    :rtype: nx.Graph
    """
    return protein.add_bond_order(G)



"""Utility functions to work with graph-level features."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
import networkx as nx
import numpy as np
import pandas as pd


def convert_graph_dict_feat_to_series(
    G: nx.Graph, feature_name: str
) -> nx.Graph:
    """
    Takes in a graph and a graph-level ``feature_name``. Converts this feature to a ``pd.Series``.
    This is useful as some features are output as dictionaries and we wish to standardise this.

    :param G:  nx.Graph containing ``G.graph[f"{feature_name}"]`` (``Dict[Any, Any]``).
    :type G: nx.Graph
    :param feature_name: Name of feature to convert to dictionary.
    :type feature_name: str
    :return: nx.Graph containing ``G.graph[f"{feature_name}"]: pd.Series``.
    :rtype: nx.Graph
    """
    G.graph[feature_name] = pd.Series(G.graph[feature_name])
    return G


def aggregate_graph_feature_over_chains(
    G: nx.Graph, feature_name: str, aggregation_type: str
) -> nx.Graph:
    """
    Performs aggregation of a feature over the chains. E.g. sums/averages/min/max molecular weights for each chain.

    :param G: nx.Graph of protein containing chain-specific features.
    :type G: nx.Graph
    :param feature_name: Name of features to aggregate.
    :type feature_name: str
    :param aggregation_type: Type of aggregation to perform (``"min"`, ``"max"``, ``"sum"``, ``"mean"``).
    :type aggregation_type: str
    :raises NameError: If ``aggregation_type`` is not one of ``"min"`, ``"max"``, ``"sum"``, ``"mean"``.
    :return: nx.Graph of protein with a new aggregated feature ``G.graph[f"{feature_name}_{aggregation_type}"]``.
    :rtype: nx.Graph
    """
    if aggregation_type == "mean":
        func = np.mean
    elif aggregation_type == "max":
        func = np.max
    elif aggregation_type == "sum":
        func = np.sum
    elif aggregation_type == "min":
        func = np.min
    else:
        raise NameError(
            "Unsupported aggregation type. Please use mean, max, sum or min."
        )

    G.graph[f"{feature_name}_{aggregation_type}"] = func(
        [G.graph[f"{feature_name}_{c}"] for c in G.graph["chain_ids"]]
    )

    return G



import copy
import functools
import math

import pandas as pd
import pytest

from graphein.datasets import PDBManager
from graphein.protein.utils import read_fasta


@functools.lru_cache()
def init_class() -> PDBManager:
    return PDBManager("tmp/")


def assert_frame_not_equal(*args, **kwargs):
    try:
        pd.testing.assert_frame_equal(*args, **kwargs)
    except AssertionError:
        # frames are not equal
        pass
    else:
        # frames are equal
        raise AssertionError


def test_initialisation():
    manager = init_class()
    assert len(manager.df) == manager.get_num_chains()
    assert manager.get_longest_chain() > 16180
    assert manager.get_shortest_chain() < 2
    assert manager.get_num_unique_pdbs() > 201762
    assert len(manager.get_unique_pdbs()) == manager.get_num_unique_pdbs()
    assert set(manager.df.columns) == {
        "id",
        "pdb",
        "chain",
        "length",
        "molecule_type",
        "name",
        "sequence",
        "ligands",
        "source",
        "resolution",
        "experiment_type",
        "deposition_date",
        "split",
    }
    assert set(manager.get_experiment_types()) == {
        "diffraction",
        "NMR",
        "EM",
        "other",
    }


def test_sample():
    manager = copy.deepcopy(init_class())
    manager.sample(n=10, update=True)
    assert len(manager.df) == 10

    manager = copy.deepcopy(init_class())
    num_examples = len(manager.df)
    manager.sample(frac=0.01, update=True)
    assert len(manager.df) == math.ceil(num_examples * 0.01)


def test_molecule_type():
    manager = init_class()
    protein = manager.molecule_type("protein")
    assert protein.molecule_type.unique() == ["protein"]

    dna = manager.molecule_type("na")
    assert dna.molecule_type.unique() == ["na"]


def test_to_csv():
    manager = copy.deepcopy(init_class())
    manager.sample(n=10, update=True)
    manager.to_csv("tmp/out.csv")
    csv = pd.read_csv("tmp/out.csv", keep_default_na=False, na_values=["N/A"])
    pd.testing.assert_frame_equal(
        csv.reset_index(drop=True), manager.df.reset_index(drop=True)
    )


def test_reset():
    manager = copy.deepcopy(init_class())
    manager.sample(n=10, update=True)
    assert_frame_not_equal(manager.df, manager.source)
    manager.reset()
    pd.testing.assert_frame_equal(manager.df, manager.source)


def test_to_fasta():
    manager = copy.deepcopy(init_class())
    manager.sample(n=10, update=True)
    manager.to_fasta("tmp/out.fasta")

    fasta = read_fasta("tmp/out.fasta")
    for k, v in fasta.items():
        assert v == manager.df[manager.df.id == k].sequence.values[0]
    assert len(fasta) == len(manager.df)


def test_experiment_type():
    manager = init_class()
    diffraction = manager.experiment_type("diffraction")
    print(diffraction)
    assert diffraction.experiment_type.unique() == ["diffraction"]
    nmr = manager.experiment_type("NMR")
    assert nmr.experiment_type.unique() == ["NMR"]
    em = manager.experiment_type("EM")
    assert em.experiment_type.unique() == ["EM"]



"""Node featurisation utilities for Gene Regulatory Networks."""

from typing import Any, Dict

from bioservices import HGNC, UniProt
from loguru import logger as log


def add_sequence_to_nodes(n: str, d: Dict[str, Any]):
    """
    Maps UniProt ACC to UniProt ID. Retrieves sequence from UniProt and adds
    it to the node

    :param n: Graph node. Unused (retained for a consistent function signature).
    :type n: str
    :param d: Graph attribute dictionary.
    :type d: Dict[str, Any]
    """

    h = HGNC(verbose=False)
    u = UniProt(verbose=False)

    d["uniprot_ids"] = h.fetch("symbol", d["gene_id"])["response"]["docs"][0][
        "uniprot_ids"
    ]

    # Todo these API calls should probably be batched
    # Todo mapping with bioservices to support other protein IDs?

    for id in d["uniprot_ids"]:
        d[f"sequence_{id}"] = u.search(id, columns="sequence").split("\n")[1]



"""Utility functions for converting between representations of protein structures."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

from typing import List, Tuple, Union

from loguru import logger as log

from graphein.utils.dependencies import import_message

from .reconstruction import get_ideal_backbone_coords
from .types import AtomTensor, BackboneFrameTensor, BackboneTensor, CoordTensor

try:
    import torch
except ImportError:
    message = import_message(
        "graphein.protein.tensor.representation",
        "torch",
        conda_channel="pytorch",
        pip_install=True,
    )
    log.debug(message)


def get_full_atom_coords(
    atom_tensor: AtomTensor, fill_value: float = 1e-5
) -> Tuple[CoordTensor, torch.Tensor, torch.Tensor]:
    """Converts an ``AtomTensor`` to a full atom representation.

    Return tuple of coords ``(N_atoms x 3)``, residue_index ``(N_atoms)``,
    atom_type ``(N_atoms x [0-36])``


    :param atom_tensor: AtomTensor of shape
        ``(N_residues, N_atoms (default is 37), 3)``
    :type atom_tensor: graphein.protein.tensor.AtomTensor
    :param fill_value: Value used to fill missing values. Defaults to ``1e-5``.
    :return: Tuple of coords, residue_index, atom_type
    :rtype: Tuple[CoordTensor, torch.Tensor, torch.Tensor]
    """
    # Get number of atoms per residue
    filled = atom_tensor[:, :, 0] != fill_value
    nz = filled.nonzero()
    residue_index = nz[:, 0]
    atom_type = nz[:, 1]
    coords = atom_tensor.reshape(-1, 3)
    coords = coords[coords != fill_value].reshape(-1, 3)
    return coords, residue_index, atom_type


def get_c_alpha(x: AtomTensor, index: int = 1) -> CoordTensor:
    """Returns tensor of C-alpha atoms: ``(L x 3)``

    :param x: Tensor of atom positions of shape:
        ``(N_residues, N_atoms (default=37), 3)``
    :type x: graphein.protein.tensor.types.AtomTensor
    :param index: Index of C-alpha atom in dimension 1 of the AtomTensor.
    :type index: int

    .. seealso:: :func:`get_backbone`
    """
    return x if x.ndim == 2 else x[:, index, :]


def get_backbone(
    x: AtomTensor, backbone_indices: List[int] = [0, 1, 2, 3]
) -> BackboneTensor:
    """Returns tensor of backbone atoms: ``(L x 4 x 3)``

    :param x: AtomTensor to extract backbone from
    :type x: AtomTensor
    :param backbone_indices: List of indices of ``[N, Ca, C, O]`` atoms (in
        order).
    :return: Tensor of backbone atoms.
    :rtype: BackboneTensor
    """
    indices = torch.tensor(backbone_indices, device=x.device)
    return x[:, indices, :]


def coarsen_sidechain(
    x: AtomTensor,
    backbone_indices: List[int] = [0, 1, 2, 3],
    aggregation: str = "mean",
) -> CoordTensor:
    """Returns tensor of sidechain centroids: ``(L x 3``

    :param x: Tensor of atom positions of shape
        ``(N_residues, N_atoms (default=37), 3)``
    :type x: graphein.protein.tensor.AtomTensor
    :param backbone_indices: List of indices in dimension 1 of the AtomTensor
        that correspond to backbone atoms (N, Ca, C, O). Defaults to
        ``[0, 1, 2, 3]``.
    :type backbone_indices: List[int]
    :param aggregation: Aggregation method to use. Defaults to ``"mean"``.
    :type aggregation: str
    :return: Tensor of sidechain centroids of shape ``(N_residues, 3)``
    :rtype: graphein.protein.tensor.CoordTensor
    :raises NotImplementedError: If aggregation method is not implemented.
    """
    # sourcery skip: remove-unnecessary-else, swap-if-else-branches
    # Select indices
    sidechain_indices = [
        a for a in range(x.shape[1]) if a not in backbone_indices
    ]
    sidechain_points = x[:, sidechain_indices, :]

    # Compute mean sidechain position
    if aggregation == "mean":
        sidechain_points = torch.mean(sidechain_points, dim=1)
        # TODO mask filled positions
    else:
        raise NotImplementedError(
            f"Aggregation method {aggregation} not implemented."
        )

    return sidechain_points


def get_backbone_frames(
    x: Union[AtomTensor, BackboneTensor],
    bb_indices: List[int] = [0, 1, 2, 4],
) -> Tuple[BackboneFrameTensor, CoordTensor]:
    """
    Return the backbone frames from the atom tensor (tuple of rotation matrices
    and alpha carbon positions).

    .. note:: ``bb_indices`` should be in the order ``[N, Ca, C, Cb]``

    :param x: Structure to get backbone rotation frames for
    :type x: Union[AtomTensor, BackboneTensor]
    :return: Tuple of RotationMatrices and Translations
    """
    from .geometry import kabsch

    # Get idealised residues centered on CA at origin
    bb = get_ideal_backbone_coords(x.shape[0], ca_center=True, device=x.device)
    indices = torch.tensor(bb_indices, device=x.device)

    return kabsch(
        bb, x[:, indices, :], residue_wise=True, return_transformed=False
    )



"""Tests for graphein.rna.subgraphs"""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

from pathlib import Path

import networkx as nx
import numpy as np

from graphein.rna.graphs import construct_graph
from graphein.rna.subgraphs import (
    extract_k_hop_subgraph,
    extract_subgraph_by_bond_type,
    extract_subgraph_by_sequence_position,
    extract_subgraph_from_atom_types,
    extract_subgraph_from_chains,
    extract_subgraph_from_node_list,
    extract_subgraph_from_point,
    extract_subgraph_from_residue_types,
)


def test_node_list_subgraphing():
    """Tests subgraph extraction from a list of nodes."""
    file_path = Path(__file__).parent / "test_data/2jyf.pdb"
    NODE_LIST = [
        "A:A:4:121:N3",
        "A:A:4:122:C4",
        "A:U:5:134:P",
        "A:U:5:135:OP1",
    ]

    G = construct_graph(path=str(file_path))

    g = extract_subgraph_from_node_list(G, NODE_LIST, filter_dataframe=True)

    # Check we get back a graph and it contains the correct nodes
    assert isinstance(g, nx.Graph)
    assert len(g) == len(NODE_LIST)
    for n in g.nodes():
        assert n in NODE_LIST
    assert (
        g.graph["pdb_df"]["node_id"]
        .str.contains("|".join(NODE_LIST), case=True)
        .all()
    )

    # Check the list of nodes is the same as the list of nodes in the original graph
    returned_node_list = extract_subgraph_from_node_list(
        G, NODE_LIST, return_node_list=True
    )
    assert all(elem in NODE_LIST for elem in returned_node_list)

    # Check there is no overlap when we inverse the selection
    g = extract_subgraph_from_node_list(
        G, NODE_LIST, inverse=True, filter_dataframe=True
    )
    assert len(g) == len(G) - len(NODE_LIST)
    for n in g.nodes():
        assert n not in NODE_LIST

    assert not (
        g.graph["pdb_df"]["node_id"]
        .str.contains("|".join(NODE_LIST), case=True)
        .any()
    )

    returned_node_list = extract_subgraph_from_node_list(
        G, NODE_LIST, inverse=True, return_node_list=True
    )

    assert all(elem not in NODE_LIST for elem in returned_node_list)


def test_extract_subgraph_from_atom_types():
    """Tests subgraph extraction from a list of allowed atom types"""
    file_path = Path(__file__).parent / "test_data/2jyf.pdb"
    G = construct_graph(path=str(file_path))

    ATOM_TYPES = ["P"]
    g = extract_subgraph_from_atom_types(G, ATOM_TYPES, filter_dataframe=True)
    assert isinstance(g, nx.Graph)
    assert len(g) == 84


def test_extract_subgraph_from_residue_types():
    """Tests subgraph extraction from a list of nodes."""
    file_path = Path(__file__).parent / "test_data/2jyf.pdb"
    RESIDUE_TYPES = ["A", "G"]

    A_COUNT = 484  # TODO
    G_COUNT = 592  # TODO

    G = construct_graph(path=str(file_path))

    g = extract_subgraph_from_residue_types(
        G, RESIDUE_TYPES, filter_dataframe=True
    )

    # Check we get back a graph and it contains the correct nodes
    assert isinstance(g, nx.Graph)
    assert len(g) == A_COUNT + G_COUNT
    for n, d in g.nodes(data=True):
        assert d["residue_name"] in RESIDUE_TYPES
    assert (
        g.graph["pdb_df"]["residue_name"]
        .str.contains("|".join(RESIDUE_TYPES), case=True)
        .all()
    )

    assert (
        len([n for n, d in g.nodes(data=True) if d["residue_name"] == "A"])
        == A_COUNT
    )
    assert (
        len([n for n, d in g.nodes(data=True) if d["residue_name"] == "G"])
        == G_COUNT
    )

    # Check the list of nodes is the same as the list of nodes in the original graph
    returned_node_list = extract_subgraph_from_node_list(
        G, RESIDUE_TYPES, return_node_list=True
    )
    assert all(elem in RESIDUE_TYPES for elem in returned_node_list)

    # Check there is no overlap when we inverse the selection
    g = extract_subgraph_from_residue_types(
        G, RESIDUE_TYPES, inverse=True, filter_dataframe=True
    )

    # assert len(g) == (len(G) - GLYCINES - ALANINES - SERINES)
    for n in g.nodes():
        assert n not in RESIDUE_TYPES

    assert not (
        g.graph["pdb_df"]["residue_name"]
        .str.contains("|".join(RESIDUE_TYPES), case=True)
        .any()
    )

    returned_node_list = extract_subgraph_from_residue_types(
        G, RESIDUE_TYPES, inverse=True, return_node_list=True
    )

    assert all(elem not in RESIDUE_TYPES for elem in returned_node_list)


def test_extract_subgraph_from_point():
    """Tests subgraph extraction from a spherical selection."""
    file_path = Path(__file__).parent / "test_data/2jyf.pdb"
    G = construct_graph(path=str(file_path))

    POINT = np.array([0.0, 0.0, 0.0])
    RADIUS = 10
    s_g = extract_subgraph_from_point(G, POINT, RADIUS, filter_dataframe=True)

    # Check all nodes are within the sphere
    for n, d in s_g.nodes(data=True):
        assert np.linalg.norm(d["coords"] - POINT) < RADIUS

    # Check we have extracted all the nodes
    for n, d in G.nodes(data=True):
        if np.linalg.norm(d["coords"] - POINT) < RADIUS:
            assert n in s_g.nodes()

    s_g = extract_subgraph_from_point(
        G, POINT, RADIUS, filter_dataframe=True, inverse=True
    )

    # Check all nodes are not within the sphere
    for n, d in s_g.nodes(data=True):
        assert np.linalg.norm(d["coords"] - POINT) > RADIUS

    # Check we have extracted all the nodes
    for n, d in G.nodes(data=True):
        if np.linalg.norm(d["coords"] - POINT) > RADIUS:
            assert n in s_g.nodes()


def test_extract_subgraph_from_chains():
    """Tests subgraph extraction from chains."""
    file_path = Path(__file__).parent / "test_data/2jyf.pdb"
    G = construct_graph(path=str(file_path))

    CHAINS = ["A", "C"]
    s_g = extract_subgraph_from_chains(G, CHAINS, filter_dataframe=True)

    # Test we only selected the correct chains
    for n, d in s_g.nodes(data=True):
        assert d["chain_id"] in CHAINS

    # Test we have extracted all the nodes
    for n, d in G.nodes(data=True):
        if d["chain_id"] in CHAINS:
            assert n in s_g.nodes()

    # Test the dataframe is correct
    assert s_g.graph["pdb_df"]["chain_id"].isin(CHAINS).all()

    s_g = extract_subgraph_from_chains(
        G, CHAINS, filter_dataframe=True, inverse=True
    )

    # Test we only selected the correct chains
    for n, d in s_g.nodes(data=True):
        assert d["chain_id"] not in CHAINS

    # Test we have extracted all the nodes
    for n, d in G.nodes(data=True):
        if d["chain_id"] in CHAINS:
            assert n not in s_g.nodes()


# @pytest.mark.skip(reason="TODO")
def test_extract_subgraph_from_sequence_position():
    """Tests subgraph extraction from sequence position."""
    file_path = Path(__file__).parent / "test_data/2jyf.pdb"
    G = construct_graph(path=str(file_path))

    SEQ_POS = list(range(1, 50, 2))

    s_g = extract_subgraph_by_sequence_position(
        G,
        SEQ_POS,
        filter_dataframe=True,
    )
    # Test we only selected the correct chains
    for n, d in s_g.nodes(data=True):
        assert d["residue_number"] in SEQ_POS

    # Test we have extracted all the nodes
    for n, d in G.nodes(data=True):
        if d["residue_number"] in SEQ_POS:
            assert n in s_g.nodes()

    # Test the dataframe is correct
    assert s_g.graph["pdb_df"]["residue_number"].isin(SEQ_POS).all()

    s_g = extract_subgraph_by_sequence_position(
        G, SEQ_POS, filter_dataframe=True, inverse=True
    )
    # Test we only selected the correct chains
    for n, d in s_g.nodes(data=True):
        assert d["residue_number"] not in SEQ_POS

    # Test we have extracted all the nodes
    for n, d in G.nodes(data=True):
        if d["residue_number"] in SEQ_POS:
            assert n not in s_g.nodes()


def test_extract_subgraph_from_bond_type():
    """Tests subgraph extraction from bond type"""
    file_path = Path(__file__).parent / "test_data/2jyf.pdb"
    G = construct_graph(path=str(file_path))

    BOND_TYPES = ["covalent"]

    s_g = extract_subgraph_by_bond_type(G, BOND_TYPES, filter_dataframe=True)

    for u, v, d in G.edges(data=True):
        if d["kind"] in BOND_TYPES:
            assert u in s_g.nodes()
            assert v in s_g.nodes()
            assert (u, v) in s_g.edges()

    for u, v, d in s_g.edges(data=True):
        for bond in list(d["kind"]):
            assert bond in BOND_TYPES

    s_g = extract_subgraph_by_bond_type(
        G, BOND_TYPES, filter_dataframe=True, inverse=True
    )

    for u, v, d in G.edges(data=True):
        if d["kind"] in BOND_TYPES:
            assert (u, v) not in s_g.edges()

    for u, v, d in s_g.edges(data=True):
        for bond in list(d["kind"]):
            assert bond not in BOND_TYPES


def test_extract_k_hop_subgraph():
    """Tests k-hop subgraph extraction."""
    file_path = Path(__file__).parent / "test_data/2jyf.pdb"
    G = construct_graph(path=str(file_path))

    CENTRAL_NODE = "A:A:6:181:N6"
    K = 1
    s_g = extract_k_hop_subgraph(G, CENTRAL_NODE, K, filter_dataframe=True)

    for n in s_g.nodes():
        if n != CENTRAL_NODE:
            assert n in list(G.neighbors(CENTRAL_NODE))

    for n in list(G.neighbors(CENTRAL_NODE)):
        assert n in s_g.nodes()



"""Tests for graphein.protein.graphs"""

from functools import partial
from pathlib import Path

import networkx as nx
import numpy as np
import pandas as pd
import pytest

from graphein.protein.config import DSSPConfig, ProteinGraphConfig
from graphein.protein.edges.distance import (
    add_aromatic_interactions,
    add_aromatic_sulphur_interactions,
    add_cation_pi_interactions,
    add_delaunay_triangulation,
    add_distance_threshold,
    add_disulfide_interactions,
    add_hydrogen_bond_interactions,
    add_hydrophobic_interactions,
    add_ionic_interactions,
    add_k_nn_edges,
    add_peptide_bonds,
)
from graphein.protein.features.nodes.aaindex import aaindex1
from graphein.protein.features.nodes.amino_acid import (
    expasy_protein_scale,
    meiler_embedding,
)
from graphein.protein.features.nodes.dssp import (
    asa,
    phi,
    psi,
    rsa,
    secondary_structure,
)
from graphein.protein.features.sequence.embeddings import (
    biovec_sequence_embedding,
    esm_residue_embedding,
    esm_sequence_embedding,
)
from graphein.protein.features.sequence.sequence import molecular_weight
from graphein.protein.graphs import (
    compute_chain_graph,
    compute_secondary_structure_graph,
    construct_graph,
    construct_graphs_mp,
    deprotonate_structure,
    read_pdb_to_dataframe,
)
from graphein.utils.dependencies import is_tool

PDB_DATA_PATH = Path(__file__).resolve().parent / "test_data" / "4hhb.pdb"
CIF_DATA_PATH = Path(__file__).resolve().parent / "test_data" / "4hhb.cif"

DSSP_AVAILABLE = is_tool("mkdssp")


def generate_graph():
    """Generate PDB network.
    This is a helper function.
    """
    return construct_graph(path=str(PDB_DATA_PATH))


@pytest.fixture(scope="module")
def net():
    """Generate proteingraph from 4HHB.pdb."""
    return generate_graph()


@pytest.fixture()
def pdb_df():
    """Generate pdb_df from 4HHB.pdb."""
    return read_pdb_to_dataframe(PDB_DATA_PATH)


@pytest.fixture()
def cif_to_pdb_df():
    """Generate pdb_df from 4HHB.cif."""
    return read_pdb_to_dataframe(CIF_DATA_PATH)


def test_nodes_are_strings(net):
    """
    Checks to make sure that the nodes are a string.
    For expediency, checks only 1/4 of the nodes.
    """
    for n in net.nodes():
        assert isinstance(n, str)


def test_pdb_vs_cif_file_parsing():
    """Generate graph from cif and pdb file and compare them"""
    G_pdb = construct_graph(path=str(PDB_DATA_PATH))
    G_cif = construct_graph(path=str(CIF_DATA_PATH))
    assert len(G_cif.nodes()) == len(G_pdb.nodes())


# Example-based Graph Construction test
def test_construct_graph():
    """Example-based test that graph construction works correctly.

    Uses 4hhb PDB file as an example test case.
    """
    G = construct_graph(path=str(PDB_DATA_PATH))
    assert isinstance(G, nx.Graph)
    assert len(G) == 574

    # Check number of peptide bonds
    peptide_bond_edges = [
        (u, v)
        for u, v, d in G.edges(data=True)
        if d["kind"] == {"peptide_bond"}
    ]
    assert len(peptide_bond_edges) == 570


@pytest.mark.skipif(not DSSP_AVAILABLE, reason="DSSP not installed.")
def test_construct_graph_with_dssp():
    """Makes sure protein graphs can be constructed with dssp

    Uses uses both a pdb code (6YC3) and a local pdb file to do so.
    """
    dssp_config_functions = {
        "edge_construction_functions": [
            add_peptide_bonds,
            add_aromatic_interactions,
            add_hydrogen_bond_interactions,
            add_disulfide_interactions,
            add_ionic_interactions,
            add_aromatic_sulphur_interactions,
            add_cation_pi_interactions,
        ],
        "graph_metadata_functions": [asa, rsa],
        "node_metadata_functions": [
            meiler_embedding,
            partial(expasy_protein_scale, add_separate=True),
        ],
        "dssp_config": DSSPConfig(),
    }

    dssp_prot_config = ProteinGraphConfig(**dssp_config_functions)

    g_pdb = construct_graph(
        config=dssp_prot_config, pdb_code="6yc3"
    )  # should download 6yc3.pdb to pdb_dir

    assert g_pdb.graph["pdb_code"] == "6yc3"
    assert g_pdb.graph["path"] is None
    assert g_pdb.graph["name"] == g_pdb.graph["pdb_code"]
    assert len(g_pdb.graph["dssp_df"]) == 1365

    file_path = str(
        Path(__file__).parent / "test_data" / "alphafold_structure.pdb"
    )
    g_local = construct_graph(config=dssp_prot_config, path=file_path)

    assert g_local.graph["pdb_code"] is None
    assert g_local.graph["path"] == file_path
    assert g_local.graph["name"] == "alphafold_structure"
    assert len(g_local.graph["dssp_df"]) == 382


def test_construct_graphs_mp():
    graph_list = [
        "2olg",
        "1bjq",
        "1omr",
        "1a4g",
        "2je9",
        "3vm5",
        "1el1",
        "3fzo",
        "1mn1",
        "1ff5",
        "1fic",
        "3a47",
        "1bir",
    ] * 5

    g = construct_graphs_mp(
        pdb_code_it=graph_list, config=ProteinGraphConfig(), return_dict=True
    )
    assert isinstance(g, dict)
    assert len(g.keys()) == len(graph_list) / 5
    for k, v in g.items():
        assert isinstance(v, (nx.Graph, None))
    g = construct_graphs_mp(
        pdb_code_it=graph_list, config=ProteinGraphConfig(), return_dict=False
    )
    assert isinstance(g, list)
    assert len(g) == len(graph_list)


def test_chain_selection():
    """Example-based test that chain selection works correctly.

    Uses 4hhb PDB file as an example test case.
    """
    file_path = Path(__file__).parent / "test_data" / "4hhb.pdb"
    G = construct_graph(path=str(file_path))

    # Check default construction contains all chains
    assert G.graph["chain_ids"] == ["A", "B", "C", "D"]
    # Check nodes contain residues from chains
    for n, d in G.nodes(data=True):
        assert d["chain_id"] in ["A", "B", "C", "D"]

    # Check graph contains only chain selection
    G = construct_graph(path=str(file_path), chain_selection=["A", "D"])
    assert G.graph["chain_ids"] == ["A", "D"]
    # Check nodes only contain residues from chain selection
    for n, d in G.nodes(data=True):
        assert d["chain_id"] in ["A", "D"]


# Edge construction tests
# Removed - testing with GetContacts as a dependency is not a priority right now
"""
def test_intramolecular_edges():
    Example-based test that intramolecular edge construction using GetContacts
    works correctly.

    Uses 4hhb PDB file as an example test case.

    file_path = Path(__file__).parent / "test_data/4hhb.pdb"

    edge_functions = {
        "edge_construction_functions": [
            hydrogen_bond,
            hydrophobic,
            peptide_bonds,
            pi_cation,
            pi_stacking,
            salt_bridge,
            t_stacking,
            van_der_waals,
        ]
    }
    config = ProteinGraphConfig(**edge_functions)
    G = construct_graph(path=str(file_path), config=config)
    # Todo complete
"""


def test_distance_edges():
    """Example-based test that distance-based edge construction works correctly,
    and that edges fulfill the distance threshold condition.

    Uses 4hhb and P04629 PDB file as an example test case.
    """
    file_path = Path(__file__).parent / "test_data" / "4hhb.pdb"

    edge_functions = {
        "edge_construction_functions": [
            partial(add_k_nn_edges, k=5, long_interaction_threshold=10),
            add_hydrophobic_interactions,
            # Todo removed for now as ring centroids require precomputing
            add_aromatic_interactions,
            add_aromatic_sulphur_interactions,
            add_delaunay_triangulation,
            add_cation_pi_interactions,
            add_peptide_bonds,
            add_hydrogen_bond_interactions,
            add_disulfide_interactions,
            add_ionic_interactions,
            partial(
                add_distance_threshold,
                threshold=12,
                long_interaction_threshold=10,
            ),
        ]
    }
    config = ProteinGraphConfig(**edge_functions)
    G1 = construct_graph(path=str(file_path), config=config)
    assert G1 is not None

    """Use AlphaFold structure."""
    file_path = Path(__file__).parent / "test_data" / "P04629.pdb"
    threshold_distance = 10.0
    new_edge_funcs = {
        "edge_construction_functions": [
            partial(
                add_distance_threshold,
                long_interaction_threshold=-1,
                threshold=threshold_distance,
            )
        ]
    }
    config = ProteinGraphConfig(**new_edge_funcs)
    G2 = construct_graph(
        path=file_path,
        config=config,
    )
    assert G2 is not None
    # Check that the distance values of each edge are less than the threshold
    for u, v, d in G2.edges(data=True):
        assert d["distance"] < threshold_distance


# Featurisation tests
@pytest.mark.skipif(not DSSP_AVAILABLE, reason="DSSP not installed.")
def test_node_features():
    # Todo this test requires attention
    # Tests node featurisers for a residue graph:
    # Amino acid features, ESM embedding, DSSP features, aaindex features

    file_path = Path(__file__).parent / "test_data" / "4hhb.pdb"

    config_params = {
        "node_metadata_functions": [
            expasy_protein_scale,  # Todo we need to refactor node data
            # assignment flow
            meiler_embedding,
        ],
        "graph_metadata_functions": [
            rsa,
            asa,
            phi,
            psi,
            secondary_structure,
            # partial(aaindex1, accession="FAUJ880111"),
        ],
        "dssp_config": DSSPConfig(),
    }
    config = ProteinGraphConfig(**config_params)
    G = construct_graph(path=str(file_path), config=config)

    # Check for existence of features
    for _, d in G.nodes(data=True):
        # Todo these functions return pd.Series, rather than adding to the node
        assert "meiler" in d.keys()
        assert "expasy" in d.keys()
        assert "rsa" in d.keys()
        assert "asa" in d.keys()
        assert "phi" in d.keys()
        assert "psi" in d.keys()
        assert "ss" in d.keys()
        continue


@pytest.mark.skip(reason="Pretrained model download is large.")
def test_sequence_features():
    # Tests sequence featurisers for a residue graph:
    # ESM and BioVec embeddings, propy and sequence descriptors
    file_path = Path(__file__).parent / "test_data" / "4hhb.pdb"

    sequence_feature_functions = {
        "graph_metadata_functions": [
            # esm_sequence_embedding,
            # esm_residue_embedding,
            biovec_sequence_embedding,
            molecular_weight,
        ]
    }
    config = ProteinGraphConfig(**sequence_feature_functions)
    G = construct_graph(path=str(file_path), config=config)

    # Check for existence on sequence-based features as node-level features
    # for n, d in G.nodes(data=True):
    # Todo this can probably be improved.
    # This only checks for the existence and shape of the esm_embedding for each
    # node
    # assert "esm_embedding" in d
    # assert len(d["esm_embedding"]) == 1280

    # Check for existence of sequence-based features as Graph-level features
    for chain in G.graph["chain_ids"]:
        assert f"sequence_{chain}" in G.graph
        # assert f"esm_embedding_{chain}" in G.graph
        assert f"biovec_embedding_{chain}" in G.graph
        assert f"molecular_weight_{chain}" in G.graph


# Checks that the sequence is extracted correctly from PDB file
def test_graph_sequence_feature():
    pdb_dir = Path(__file__).parent / "test_data"
    pdb = "4hhb"

    g_atom = construct_graph(
        pdb_code=pdb,
        config=ProteinGraphConfig(
            pdb_dir=pdb_dir,
            granularity="atom",  # atomistic
        ),
    )
    g_res = construct_graph(
        pdb_code=pdb,
        config=ProteinGraphConfig(
            pdb_dir=pdb_dir,
            granularity="CA",  # residue
        ),
    )

    for c in g_atom.graph["chain_ids"]:
        # assert sequences are equal
        assert g_atom.graph[f"sequence_{c}"] == g_res.graph[f"sequence_{c}"]


def test_insertion_and_alt_loc_handling():
    configs = {
        "granularity": "CA",
        "keep_hets": [],
        "insertions": False,
        "alt_locs": "max_occupancy",
        "verbose": False,
        "node_metadata_functions": [meiler_embedding, expasy_protein_scale],
        "edge_construction_functions": [
            add_peptide_bonds,
            add_hydrogen_bond_interactions,
            add_ionic_interactions,
            add_aromatic_sulphur_interactions,
            add_hydrophobic_interactions,
            add_cation_pi_interactions,
        ],
    }

    config = ProteinGraphConfig(**configs)

    # This is a nasty PDB with a lot of insertions and altlocs
    g = construct_graph(config=config, pdb_code="6OGE")

    assert len(g.graph["sequence_A"]) + len(g.graph["sequence_B"]) + len(
        g.graph["sequence_C"]
    ) + len(g.graph["sequence_D"]) + len(g.graph["sequence_E"]) == len(g)
    assert g.graph["coords"].shape[0] == len(g)


def test_alt_loc_exclusion():
    configs = {
        "granularity": "CA",
        "keep_hets": [],
        "insertions": True,
        "alt_locs": "max_occupancy",
        "verbose": False,
        "node_metadata_functions": [meiler_embedding, expasy_protein_scale],
        "edge_construction_functions": [
            add_peptide_bonds,
            add_hydrogen_bond_interactions,
            add_ionic_interactions,
            add_aromatic_sulphur_interactions,
            add_hydrophobic_interactions,
            add_cation_pi_interactions,
        ],
    }

    config = ProteinGraphConfig(**configs)

    # This is a PDB with three altlocs
    g = construct_graph(config=config, pdb_code="2VVI")

    # Test altlocs are dropped
    assert len(set(g.nodes())) == len(g.nodes())

    # Test the correct one is left
    for opt, expected_coords, node_id in (
        ("max_occupancy", [5.850, -9.326, -42.884], "A:CYS:195:A"),
        ("min_occupancy", [5.864, -9.355, -42.943], "A:CYS:195:B"),
        ("first", [5.850, -9.326, -42.884], "A:CYS:195:A"),
        ("last", [5.864, -9.355, -42.943], "A:CYS:195:B"),
    ):
        config.alt_locs = opt
        g = construct_graph(config=config, pdb_code="2VVI")
        assert np.array_equal(
            g.nodes[node_id]["coords"],
            np.array(expected_coords, dtype=np.float32),
        )


def test_alt_loc_inclusion():
    configs = {
        "granularity": "CA",
        "keep_hets": [],
        "insertions": False,
        "alt_locs": True,
        "verbose": False,
        "node_metadata_functions": [meiler_embedding, expasy_protein_scale],
        "edge_construction_functions": [
            add_peptide_bonds,
            add_hydrogen_bond_interactions,
            add_ionic_interactions,
            add_aromatic_sulphur_interactions,
            add_hydrophobic_interactions,
            add_cation_pi_interactions,
        ],
    }

    config = ProteinGraphConfig(**configs)

    # This is a PDB with an altloc leading to different residues
    g = construct_graph(config=config, pdb_code="1ALX")

    # Test both are present
    assert "A:TYR:11:A" in g.nodes() and "A:TRP:11:B" in g.nodes()

    # TODO Test on other PDBs where altlocs are of the same residues


def test_edges_do_not_add_nodes_for_chain_subset():
    new_funcs = {
        "edge_construction_functions": [
            add_peptide_bonds,
            add_hydrogen_bond_interactions,
            add_disulfide_interactions,
            add_ionic_interactions,
            add_aromatic_interactions,
            add_aromatic_sulphur_interactions,
            add_cation_pi_interactions,
        ],
    }
    config = ProteinGraphConfig(**new_funcs)
    g = construct_graph(config=config, pdb_code="2vvi", chain_selection=["A"])
    assert len(g) == 217
    g = construct_graph(config=config, pdb_code="2vvi", chain_selection=["B"])
    assert len(g) == 219
    g = construct_graph(config=config, pdb_code="2vvi", chain_selection=["C"])
    assert len(g) == 222
    g = construct_graph(config=config, pdb_code="2vvi", chain_selection=["D"])
    assert len(g) == 219


@pytest.mark.skipif(not DSSP_AVAILABLE, reason="DSSP not installed.")
def test_secondary_structure_graphs():
    file_path = Path(__file__).parent / "test_data" / "4hhb.pdb"
    config = ProteinGraphConfig(
        edge_construction_functions=[
            add_hydrophobic_interactions,
            add_aromatic_interactions,
            add_disulfide_interactions,
            add_peptide_bonds,
            add_hydrogen_bond_interactions,
        ],
        graph_metadata_functions=[secondary_structure],
        dssp_config=DSSPConfig(),
    )
    g = construct_graph(path=str(file_path), config=config)

    h = compute_secondary_structure_graph(g, remove_non_ss=False)
    # Check number of residues preserved
    res_counts = sum(d["residue_counts"] for _, d in h.nodes(data=True))
    assert res_counts == len(
        g
    ), "Residue counts in SS graph should match number of residues in original \
        graph"
    assert nx.is_connected(
        h
    ), "SS graph should be connected in this configuration"

    h = compute_secondary_structure_graph(
        g,
        remove_non_ss=False,
        remove_self_loops=False,
        return_weighted_graph=False,
    )
    assert len(g.edges) == len(
        h.edges
    ), "Multigraph should have same number of edges."


@pytest.mark.skipif(not DSSP_AVAILABLE, reason="DSSP not installed.")
def test_chain_graph():
    file_path = Path(__file__).parent / "test_data" / "4hhb.pdb"
    config = ProteinGraphConfig(
        edge_construction_functions=[
            add_hydrophobic_interactions,
            add_aromatic_interactions,
            add_disulfide_interactions,
            add_peptide_bonds,
            add_hydrogen_bond_interactions,
        ],
        graph_metadata_functions=[secondary_structure],
        dssp_config=DSSPConfig(),
    )
    g = construct_graph(path=str(file_path), config=config)
    h = compute_chain_graph(g)
    assert len(h.edges) == len(g.edges), "Number of edges do not match"

    h = compute_chain_graph(g, return_weighted_graph=True)
    node_sum = sum(d["num_residues"] for _, d in h.nodes(data=True))
    assert node_sum == len(g), "Number of residues do not match"


def test_df_processing():
    def return_even_df(df: pd.DataFrame) -> pd.DataFrame:
        return df.loc[df["residue_number"] % 2 == 0]

    def remove_hetatms(df: pd.DataFrame) -> pd.DataFrame:
        return df.loc[df["record_name"] == "ATOM"]

    params_to_change = {
        "protein_df_processing_functions": [return_even_df, remove_hetatms],
        "granularity": "atom",
    }

    config = ProteinGraphConfig(**params_to_change)
    config.dict()

    config2 = ProteinGraphConfig(granularity="atom", deprotonate=True)

    g1 = construct_graph(config=config, pdb_code="3eiy")
    g2 = construct_graph(config=config2, pdb_code="3eiy")
    g3 = construct_graph(config=config2, pdb_code="4cvi")

    for n, d in g1.nodes(data=True):
        assert (
            int(d["residue_number"]) % 2 == 0
        ), "Only even residues should be present"

    assert len(g1) != len(g2), "Graphs should not be equal"
    for n, d in g3.nodes(data=True):
        assert d["element_symbol"] not in [
            "H",
            "D",
            "T",
        ], "No hydrogen isotopes should be present"

    config3 = ProteinGraphConfig(granularity="atom", deprotonate=False)
    g4 = construct_graph(config=config3, pdb_code="4cvi")
    has_H = []
    for n, d in g4.nodes(data=True):
        if d["element_symbol"] in {"H", "D", "T"}:
            has_H.append(n)
    assert len(has_H) > 0, "No hydrogen isotopes are present"



"""Functions for parsing GRNs from RegNetwork."""

# %%
# Graphein
# Author: Ramon Vinas, Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

import functools
import os
import ssl
import zipfile
from pathlib import Path
from typing import Callable, List, Optional

import pandas as pd
import wget
from loguru import logger as log

from graphein.utils.utils import deprecated, filter_dataframe, ping


@deprecated(
    "RegNetwork appears to be down. This warning will be removed in a future release if the service is restored."
)
def _download_RegNetwork(
    root_dir: Optional[Path] = None, network_type: str = "human"
) -> str:
    """
    Downloads RegNetwork regulatory interactions to the root directory.
    Returns the filepath.

    :param root_dir: ``Path`` object specifying the location to download
        RegNetwork to. Default is ``None`` which downloads to the ``dataset/``
        directory inside graphein.
    :type root_dir: pathlib.Path, optional
    :param network_type: Specifies whether to download human or mouse
        regulatory network. Supported values: ``"human"`` (default),
        ``"mouse"``.
    :type network_type: str
    :returns: path to downloaded RegNetwork
    :rtype: str
    """

    # Ping server to check if file is available
    ping_result = ping("regnetworkweb.org")
    if not ping_result:
        log.warning("RegNetwork is not available. Please check your internet \
            connection or verify at: http://www.regnetworkweb.org")

    mouse_url = "https://regnetworkweb.org/download/mouse.zip"

    if network_type == "human":
        human_url = "https://regnetworkweb.org/download/human.zip"
        url = human_url
    elif network_type == "mouse":
        url = mouse_url
    else:
        raise ValueError(
            f"network_type: {network_type} is unsupported. Please use 'human' \
            or 'mouse'"
        )

    # If no root dir is provided, use the dataset directory inside graphein.
    if root_dir is None:
        root_dir = Path(__file__).parent.parent.parent / "datasets"

    regnetwork_dir = f"{root_dir}/regnetwork"
    Path(regnetwork_dir).mkdir(parents=False, exist_ok=True)
    compressed_file = f"{regnetwork_dir}/human.zip"
    out_dir = f"{regnetwork_dir}/human"
    file = f"{out_dir}/human.source"

    # Download data and unzip
    if not os.path.exists(file):
        log.info("Downloading RegNetwork ...")
        # switch ssl context for unverified download
        default_https_context = ssl._create_default_https_context
        ssl._create_default_https_context = ssl._create_unverified_context
        wget.download(url, compressed_file)
        # switch ssl context back to default
        ssl._create_default_https_context = default_https_context
        with zipfile.ZipFile(compressed_file, "r") as zip_ref:
            zip_ref.extractall(out_dir)

    return file


@deprecated(
    "RegNetwork appears to be down. This warning will be removed in a future release if the service is restored."
)
def _download_RegNetwork_regtypes(root_dir: Optional[Path] = None) -> str:
    """
    Downloads RegNetwork regulatory interactions types to the root directory.

    :param root_dir: ``Path`` object specifying the location to download
        RegNetwork to
    :type root_dir: pathlib.Path, optional. Default is ``None`` which downloads
        to the ``dataset/`` directory inside graphein.
    :returns: Path to downloaded RegNetwork
    :rtype: str
    """
    url = "https://regnetworkweb.org/download/RegulatoryDirections.zip"

    if root_dir is None:
        root_dir = Path(__file__).parent.parent.parent / "datasets"

    regnetwork_dir = f"{root_dir}/regnetwork"
    Path(regnetwork_dir).mkdir(parents=False, exist_ok=True)
    compressed_file = f"{regnetwork_dir}/RegulatoryDirections.zip"
    out_dir = f"{regnetwork_dir}/human"
    file = f"{out_dir}/new_kegg.human.reg.direction.txt"

    # Download data and unzip
    if not os.path.exists(file):
        log.info("Downloading RegNetwork reg types ...")
        # switch ssl context for unverified download
        default_https_context = ssl._create_default_https_context
        ssl._create_default_https_context = ssl._create_unverified_context
        wget.download(url, compressed_file)
        # switch ssl context back to default
        ssl._create_default_https_context = default_https_context

        with zipfile.ZipFile(compressed_file, "r") as zip_ref:
            zip_ref.extractall(out_dir)

    return file


@deprecated(
    "RegNetwork appears to be down. This warning will be removed in a future release if the service is restored."
)
@functools.lru_cache()
def load_RegNetwork_interactions(
    root_dir: Optional[Path] = None,
) -> pd.DataFrame:
    """
    Loads RegNetwork interaction datafile. Downloads the file first if not
    already present.

    :param root_dir: ``Path`` object specifying the location of the
        RegNetwork download.
    :type root_dir: pathlib.Path, optional. Default is ``None``.
    :returns: RegNetwork interaction data as a pd.DataFrame.
    :rtype: pd.DataFrame
    """
    file = _download_RegNetwork(root_dir)
    return pd.read_csv(
        file, delimiter="\t", header=None, names=["g1", "id1", "g2", "id2"]
    )


@deprecated(
    "RegNetwork appears to be down. This warning will be removed in a future release if the service is restored."
)
@functools.lru_cache()
def load_RegNetwork_regulation_types(
    root_dir: Optional[Path] = None,
) -> pd.DataFrame:
    """
    Loads RegNetwork regulation types. Downloads the file first if not already
        present.

    :param root_dir: ``Path`` object specifying the location of the RegNetwork
        download.
    :type root_dir: pathlib.Path, optional. Default is ``None``.
    :returns: RegNetwork regulation types as a pd.DataFrame.
    :rtype: pd.DataFrame
    """
    file = _download_RegNetwork_regtypes(root_dir)
    return pd.read_csv(
        file,
        delimiter=" ",
        header=None,
        names=["tf", "id1", "target", "id2", "regtype"],
        skiprows=1,
    )


@deprecated(
    "RegNetwork appears to be down. This warning will be removed in a future release if the service is restored."
)
def parse_RegNetwork(
    gene_list: List[str], root_dir: Optional[Path] = None
) -> pd.DataFrame:
    """
    Parser for RegNetwork interactions

    :param gene_list: List of gene identifiers
    :type gene_list: List[str]
    :param root_dir: ``Path`` object specifying the location of the RegNetwork
        download.
    :return: Pandas DataFrame with the regulatory interactions between genes
        in the gene list.
    :rtype: pd.DataFrame
    """
    # Load dataframes
    df = load_RegNetwork_interactions(root_dir)
    reg_df = load_RegNetwork_regulation_types(root_dir)

    df = pd.merge(
        df,
        reg_df,
        how="outer",
        left_on=["g1", "g2"],
        right_on=["tf", "target"],
    )
    df["g1"] = df["g1"].combine_first(df["tf"])
    df["g2"] = df["g2"].combine_first(df["target"])

    # Select input genes
    df = df[df["g1"].isin(gene_list) & df["g2"].isin(gene_list)]

    return df


def filter_RegNetwork(
    df: pd.DataFrame, funcs: Optional[List[Callable]] = None
) -> pd.DataFrame:
    """
    Filters results of RegNetwork call by providing a list of user-defined
    functions that accept a DataFrame and return a DataFrame

    :param df: pd.DataFrame to filter
    :param funcs: List of functions that carry out DataFrame processing.
    :return: Processed DataFrame.
    :rtype: pd.DataFrame
    """
    if funcs is not None:
        df = filter_dataframe(df, funcs)

    return df


def standardise_RegNetwork(df: pd.DataFrame) -> pd.DataFrame:
    """
    Standardises RegNetwork dataframe, e.g. puts everything into a common format

    :param df: Source specific Pandas dataframe
    :type df: pd.DataFrame
    :return: Standardised dataframe
    :rtype: pd.DataFrame
    """
    # Rename & delete columns
    df = df[["g1", "g2", "regtype"]]

    # Add source column
    df["source"] = "RegNetwork"

    # Standardise regulatory types
    df["regtype"].replace(
        {"-->": "+", "--|": "-", None: "?", "-p": "?", "-/-": "?"},
        inplace=True,
    )

    return df


@deprecated(
    "RegNetwork appears to be down. This warning will be removed in a future release if the service is restored."
)
def RegNetwork_df(
    gene_list: List[str],
    root_dir: Optional[Path] = None,
    filtering_funcs: Optional[List[Callable]] = None,
) -> pd.DataFrame:
    """
    Generates standardised dataframe with RegNetwork protein-protein
    interactions, filtered according to user's input.

    :param gene_list: List of gene identifiers.
    :type gene_list: List[str]
    :param root_dir: ``Path`` object specifying the location of the RegNetwork
        dataset.
    :type root_dir: pathlib.Path, optional. Default is ``None``.
    :param filtering_funcs: List of (optional) functions that accept a dataframe
        and return a dataframe. These functions are applied to the dataframe to
        filter entries according to the user's input.
    :type filtering_funcs: List[Callable], optional. Default is ``None``.
    :return: Standardised dataframe with RegNetwork interactions.
    :rtype: pd.DataFrame
    """
    df = parse_RegNetwork(gene_list=gene_list, root_dir=root_dir)
    df = filter_RegNetwork(df, filtering_funcs)
    df = standardise_RegNetwork(df)

    return df



"""Tests for graph format conversion procedures."""

from functools import partial

import pytest
import torch

from graphein.ml import GraphFormatConvertor
from graphein.protein.config import ProteinGraphConfig
from graphein.protein.edges.distance import add_k_nn_edges
from graphein.protein.features.nodes.amino_acid import amino_acid_one_hot
from graphein.protein.graphs import construct_graph

try:
    import torch_geometric
    from torch_geometric.utils import is_undirected

    PYG_AVAIL = True
except ImportError:
    PYG_AVAIL = False


@pytest.mark.skipif(not PYG_AVAIL, reason="PyG not installed")
@pytest.mark.parametrize("pdb_code", ["10gs", "1bui", "1cw3"])
def test_nx_to_pyg(pdb_code):
    # Construct graph of a multimer protein complex
    edge_funcs = {
        "edge_construction_functions": [
            partial(
                add_k_nn_edges,
                k=1,
                long_interaction_threshold=0,
                exclude_edges=["inter"],
                kind_name="intra",
            ),
            partial(
                add_k_nn_edges,
                k=1,
                long_interaction_threshold=0,
                exclude_edges=["intra"],
                kind_name="inter",
            ),
        ]
    }
    node_feature_funcs = {"node_metadata_functions": [amino_acid_one_hot]}
    config = ProteinGraphConfig(**edge_funcs, **node_feature_funcs)
    g = construct_graph(config=config, pdb_code=pdb_code)

    # Convert to PyG
    convertor = GraphFormatConvertor(
        src_format="nx",
        dst_format="pyg",
        columns=[
            "coords",
            "node_id",
            "amino_acid_one_hot",
            "edge_index_inter",
            "edge_index_intra",
            "edge_index",
        ],
    )
    data = convertor(g)

    # Test
    # Nodes
    assert len(data.node_id) == data.num_nodes

    # Coordinates
    assert isinstance(data.coords, torch.Tensor)
    assert data.coords.shape == torch.Size([data.num_nodes, 3])

    # Features
    assert isinstance(data.amino_acid_one_hot, torch.Tensor)
    assert data.amino_acid_one_hot.shape == torch.Size([data.num_nodes, 20])

    # Edges
    assert isinstance(data.edge_index, torch.Tensor)
    assert data.edge_index.shape[0] == 2
    assert isinstance(data.edge_index_inter, torch.Tensor)
    assert data.edge_index_inter.shape[0] == 2
    assert isinstance(data.edge_index_intra, torch.Tensor)
    assert data.edge_index_intra.shape[0] == 2
    assert (
        data.edge_index.shape[1]
        == data.edge_index_inter.shape[1] + data.edge_index_intra.shape[1]
    )

    # Directed/undirected consistency
    assert g.is_directed() is not is_undirected(data.edge_index)



"""Graph construction utilities for Gene Regulatory Networks."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Ramon Vinas
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from typing import Callable, List, Optional

import networkx as nx
from loguru import logger as log

from graphein.grn.config import GRNGraphConfig
from graphein.utils.utils import (
    annotate_edge_metadata,
    annotate_graph_metadata,
    annotate_node_metadata,
    compute_edges,
)

EDGE_COLOR_MAPPING = {"trrust": "r", "regnetwork": "b", "abasy": "g"}


def parse_kwargs_from_config(config: GRNGraphConfig) -> GRNGraphConfig:
    """
    If configs for specific dataset are provided in the Global GRNGraphConfig, we update the kwargs

    :param config: GRN graph configuration object.
    :type config: graphein.grn.GRNGraphConfig
    :return: config with updated config.kwargs
    :rtype: graphein.grn.GRNGraphConfig
    """
    if config.trrust_config.kwargs is not None:
        trrust_config_dict = {
            "TRRUST_" + k: v
            for k, v in dict(config.trrust_config.kwargs.items())
        }
        config.kwargs = config.kwargs.update(trrust_config_dict)

    if config.regnetwork_config.kwargs is not None:
        regnetwork_config_dict = {
            "RegNetwork_" + k: v
            for k, v in dict(config.regnetwork_config.kwargs.items())
        }
        config.kwargs = config.kwargs.update(regnetwork_config_dict)
    return config


def compute_grn_graph(
    gene_list: List[str],
    edge_construction_funcs: List[Callable],
    graph_annotation_funcs: Optional[List[Callable]] = None,
    node_annotation_funcs: Optional[List[Callable]] = None,
    edge_annotation_funcs: Optional[List[Callable]] = None,
    config: Optional[GRNGraphConfig] = None,
) -> nx.Graph:
    """
    Computes a Gene Regulatory Network Graph from a list of gene IDs

    :param gene_list: List of gene identifiers
    :type gene_list: List[str]
    :param edge_construction_funcs:  List of functions to construct edges with
    :type edge_construction_funcs: List[Callable]
    :param graph_annotation_funcs: List of functions functools annotate graph metadata, defaults to None
    :type graph_annotation_funcs: List[Callable], optional
    :param node_annotation_funcs: List of functions to annotate node metadata, defaults to None
    :type node_annotation_funcs: List[Callable], optional
    :param edge_annotation_funcs: List of functions to annotate edge metadata, defaults to None
    :type edge_annotation_funcs: List[Callable], optional
    :param config: Config specifying additional parameters for STRING and BIOGRID, defaults to None
    :type config: graphein.grn.GRNGraphConfig, optional
    :return: nx.Graph of PPI network
    :rtype: nx.Graph
    """

    # Load default config if none supplied
    if config is None:
        config = GRNGraphConfig()

    # Parse kwargs from config
    config = parse_kwargs_from_config(config)

    # Create *directed* graph and add genes as nodes
    G = nx.DiGraph(
        gene_list=gene_list,
        sources=[],
        # ncbi_taxon_id=config.ncbi_taxon_id,
    )
    G.add_nodes_from(gene_list)
    log.debug(f"Added {len(gene_list)} nodes to graph")

    nx.set_node_attributes(
        G,
        dict(zip(gene_list, gene_list)),
        "gene_id",
    )

    # Annotate additional graph metadata
    if graph_annotation_funcs is not None:
        G = annotate_graph_metadata(G, graph_annotation_funcs)

    # Annotate additional node metadata
    if node_annotation_funcs is not None:
        G = annotate_node_metadata(G, node_annotation_funcs)

    # Add edges
    G = compute_edges(G, edge_construction_funcs)

    # Annotate additional edge metadata
    if edge_annotation_funcs is not None:
        G = annotate_edge_metadata(G, edge_annotation_funcs)

    return G


if __name__ == "__main__":
    from functools import partial

    import matplotlib.pyplot as plt

    from graphein.grn.edges import add_regnetwork_edges, add_trrust_edges
    from graphein.grn.features.node_features import add_sequence_to_nodes

    gene_list = ["AATF", "MYC", "USF1", "SP1", "TP53", "DUSP1"]

    config = GRNGraphConfig()
    kwargs = config.kwargs

    def edge_ann_fn(u, v, d):
        if "+" in d["regtype"]:
            d["regtype"] = "+"
        elif "-" in d["regtype"]:
            d["regtype"] = "-"
        elif "?" in d["regtype"]:
            d["regtype"] = "?"

    g = compute_grn_graph(
        gene_list=gene_list,
        edge_construction_funcs=[
            partial(
                add_trrust_edges,
                trrust_filtering_funcs=config.trrust_config.filtering_functions,
            ),
            partial(
                add_regnetwork_edges,
                regnetwork_filtering_funcs=config.regnetwork_config.filtering_functions,
            ),
        ],
        node_annotation_funcs=[add_sequence_to_nodes],  # , molecular_weight],
        edge_annotation_funcs=[edge_ann_fn],
    )
    print(g.edges(data=True))

    edge_colors = [
        (
            "r"
            if g[u][v]["kind"] == {"trrust"}
            else "b" if g[u][v]["kind"] == {"regnetwork"} else "y"
        )
        for u, v in g.edges()
    ]

    print(nx.info(g))

    pos = nx.spring_layout(g)
    nx.draw(g, pos=pos, with_labels=True, edge_color=edge_colors)
    edge_labels = {(u, v): g[u][v]["regtype"] for u, v in g.edges}
    nx.draw_networkx_edge_labels(g, pos=pos, edge_labels=edge_labels)
    plt.show()
